import re
from typing import List
import logging

import markdown
from markdown import Extension
from markdown.preprocessors import Preprocessor
from toolz.curried import (
    pipe, curry, merge, map, filter,
)

from .. import common
from .. import yaml

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

card_begin_re = re.compile(
    r'\|\.+(?:\[(?P<card_type>.*)\])?\.+\|'
)
card_end_re = re.compile(
    r'\|\^*\|'
)

def defcard(lines):
    return lines

def first_nonempty(lines):
    for i, l in enumerate(lines):
        if l.strip():
            return i, l

'''
  <div class="card mb-3 mt-3 vuln-card hideable">
    <div class="card-body finding-card-body">
      {% for dt, dd in rows %}
        <div class="row">
          <div class="col-4">{{ dt }}</div>
          <div class="col-8">{{ dd }}</div>
        </div>
      {% endfor %}
    </div>
  </div>
'''

def make_card(lines):
    hi, header = first_nonempty(lines[1:])
    hi += 1
    
    yield from [
        '<div class="card mb-2 mt-2">',
        '<div class="card-header">' + markdown.markdown(header) + '</div>',
        '<div class="card-body">',
    ] + lines[hi + 2:] + [
        '</div>',
        '</div>',
    ]

class ReMatchCase:
    def __init__(self):
        self.matchers = {}
        self._default = lambda value, *a, **kw: value

    def match(self, regex):
        def decorator(func):
            self.matchers[re.compile(regex)] = func
            return func
        return decorator

    def default(self, func):
        self._default = func
        return func

    def __call__(self, value, *a, **kw):
        for regex, func in self.matchers.items():
            match = regex.match(value)
            if match:
                return func(value, match, *a, **kw)
        return self._default(value, *a, **kw)

parse_header = ReMatchCase()

@parse_header.match(r'^(?P<header>#+)\s+(?P<text>.*?)\s*$')
def header_tag_header(_, match):
    mdict = match.groupdict()
    h_tag = f'h{len(mdict["header"])}'
    text = mdict['text'].strip()
    id_text = pipe(
        text.lower().split(),
        '-'.join,
    )
    return (
        # f'<a name={id_text} />'
        f'<{h_tag} class="card-header mdcard-header">{text}</{h_tag}>'
    )

@parse_header.default
def bare_text_header(header):
    text = markdown.markdown(header.strip())
    id_text = pipe(
        header.strip().lower().split(),
        '-'.join,
    )
    return (
        # f'<a name={id_text} />'
        f'<div class="card-header mdcard-header">{text}</div>'
    )

def card_to_html(lines: List[str]):
    new_lines = lines[:]

    delete = []
    for i, line in enumerate(new_lines):
        begin = card_begin_re.match(line)
        end = card_end_re.match(line)
        if begin:
            new_lines[i] = '<div class="card mb-2 mt-2 mdcard">'
            hi, header = first_nonempty(new_lines[i + 1:])
            hi += i + 1

            for i in range(i + 1, hi):
                if not new_lines[i].strip():
                    delete.append(i)

            new_lines[hi] = (
                parse_header(header) +
                # '<div class="card-header mdcard-header">' +
                # markdown.markdown(header.strip()) +
                '<div class="card-body mdcard-body">'
            )
        elif end:
            new_lines[i] = '</div></div>'
    for i in delete[::-1]:
        new_lines.pop(i)
    return new_lines

# def card_to_html(lines: List[str]):
#     new_lines = lines[:]

#     cards = []
#     for i, line in enumerate(new_lines):
#         begin = card_begin_re.match(line)
#         end = card_end_re.match(line)
#         if begin:
#             cards.append((i, begin))
#         elif end and cards:
#             (i0, begin) = cards.pop()
#             i1 = i
#             card_type = begin.groupdict()['card_type']
#             if card_type and card_type in globals():
#                 yield from globals()[card_type](lines[i0: i1])
#             else:
#                 yield from make_card(lines[i0: i1])
#         elif not cards:
#             yield line

class CardPreprocessor(Preprocessor):
    def run(self, lines):
        return list(card_to_html(lines))
            
class BootstrapCardExtension(Extension):
    def extendMarkdown(self, md):
        md.preprocessors.register(
            CardPreprocessor(md), "bootstrapcard", 8
        )
        
def makeExtension(*args, **kwargs):
    return BootstrapCardExtension(*args, **kwargs)




import re
import logging

from markdown.util import etree
from markdown import Extension
from markdown.treeprocessors import Treeprocessor
from markdown.blockprocessors import BlockProcessor

Logger = logging.getLogger('mdx_sections')

def makeExtension(config=None):
    return DocumentSectionExtension(config)

class DocumentSectionExtension(Extension):
    def extendMarkdown(self, md, md_globals):
        self.config = {
            'section_cls': [
                'section', 'Default CSS class applied to sections'
            ],
            'heading_cls': [
                'title', 'CSS class applied to transformed h{1-6} element'
            ]
        }
        self.section_processor = DocumentSectionProcessor(
            self.getConfig('section_cls'), self.getConfig('heading_cls')
        )
        self.heading_processor = HeadingProcessor(
            md.parser, self.getConfig('section_cls')
        )

        md.treeprocessors.add(
            'document_section', self.section_processor, '_end'
        )
        md.parser.blockprocessors.add(
            'header_id', self.heading_processor, '_begin'
        )

class HeadingProcessor(BlockProcessor):
    pattern = re.compile(
        r'^(?P<depth>#+)\s+{(?P<id>[,A-Za-z0-9_-]+)}\s+(?P<text>.*)'
    )
    current_section = []

    def __init__(self, md, section_cls='section'):
        self.section_cls = section_cls
        BlockProcessor.__init__(self, md)

    def test(self, parent, block):
        if self.pattern.match(block):
            return True

    def run(self, parent, blocks):
        lines = blocks.pop(0).split('\n')

        heading = self.pattern.match(lines[0]).groupdict()
        depth = min(6, int(len(heading.get('depth'))))

        id = heading.get('id').split(',')
        if len(id) == 2:
            cls = id[1]
            id = id[0]
        else:
            id = id[0]
            cls = 'section'

        h = etree.SubElement(parent, 'h{depth}'.format(depth=depth))
        h.attrib['id'] = id
        h.attrib['class'] = cls
        h.text = heading.get('text')

        for i, line in enumerate(lines[1:]):
            blocks.insert(i, line)

class DocumentSectionProcessor(Treeprocessor):
    def __init__(self, section_cls='section', heading_cls='title'):
        self.section_cls = section_cls
        self.heading_cls = heading_cls
        Treeprocessor.__init__(self)

    def process_nodes(self, node):
        s = []
        pattern = re.compile(r'^h(\d)')

        for n in node.getchildren():
            match = pattern.match(n.tag.lower())

            if match:
                section = etree.SubElement(node, 'div')
                section.append(n)
                section.attrib['class'] = n.attrib.get(
                    'class', self.section_cls
                )

                node.remove(n)

                if 'id' in n.attrib:
                    section.attrib['id'] = n.attrib['id']
                    del n.attrib['id']

                n.attrib['class'] = self.heading_cls

                depth = int(match.group(1))
                Logger.debug(
                    'matched %s>%s#%s.%s', depth, n.tag,
                    n.attrib.get('id'), n.attrib.get('class')
                )

                contained = False
                while s:
                    Logger.debug('searching for container')
                    container, container_depth = s[-1]
                    if depth <= container_depth:
                        s.pop()
                    else:
                        contained = True
                        break

                if contained:
                    Logger.debug(
                        'using container %s>%s#%s.%s', container_depth,
                        container.tag, container.attrib.get('id'),
                        container.attrib.get('class')
                    )
                    container.append(section)
                    node.remove(section)

                else:
                    Logger.debug('no container found, appending to root')

                Logger.debug(
                    'current container %s>%s#%s.%s', depth, section.tag,
                    section.attrib.get('id'), section.attrib.get('class')
                )
                s.append((section, depth))

            else:
                Logger.debug('non-heading element %s', n.tag)
                if s:
                    container, container_depth = s[-1]
                    container.append(n)
                    node.remove(n)
                    Logger.debug(
                        'appending to container %s>%s#%s.%s', container_depth,
                        container.tag, container.attrib.get('id'),
                        container.attrib.get('class')
                    )

    def run(self, root):
        Logger.info('processing %s', root.tag)
        self.process_nodes(root)
        return root
